"""Continuum fitting module."""
import argparse
import logging

import numpy as np
from numba import njit
import fitsio
from scipy.optimize import curve_fit
from scipy.interpolate import UnivariateSpline, CubicSpline
from scipy.special import legendre

from mpi4py import MPI

from qsonic import QsonicException
from qsonic.spectrum import valid_spectra
from qsonic.mpi_utils import mpi_fnc_bcast, MPISaver
from qsonic.mathtools import (
    block_covariance_of_square,
    FastLinear1DInterp, FastCubic1DInterp,
    SubsampleCov
)


def add_picca_continuum_parser(parser=None):
    """ Adds PiccaContinuumFitter related arguments to parser. These
    arguments are grouped under 'Continuum fitting options'. All of them
    come with defaults, none are required.

    Arguments
    ---------
    parser: argparse.ArgumentParser, default: None

    Returns
    ---------
    parser: argparse.ArgumentParser
    """
    if parser is None:
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    cont_group = parser.add_argument_group('Continuum fitting options')

    cont_group.add_argument(
        "--num-iterations", type=int, default=10,
        help="Number of iterations for continuum fitting.")
    cont_group.add_argument(
        "--continuum-model", choices=["picca", "true", "input"],
        default="picca",
        help="Continuum model. *picca* fits the continuum in the forest "
        "region. *true* is the true continuum analysis for mock analysis. "
        " *input* in the input continuum analysis for all."
    )
    cont_group.add_argument(
        "--true-continuum", action="store_true",
        help="Alternative argument for true continuum analysis.")
    cont_group.add_argument(
        "--input-continuum-dir", help="Input continuum directory.")
    cont_group.add_argument(
        "--fiducial-meanflux", help="Fiducial mean flux FITS file.")
    cont_group.add_argument(
        "--fiducial-varlss", help="Fiducial var_lss FITS file.")
    cont_group.add_argument(
        "--cont-order", type=int, default=1,
        help="Order of continuum fitting polynomial.")
    cont_group.add_argument(
        "--var-fit-eta", action="store_true",
        help="Fit for noise calibration (eta).")
    cont_group.add_argument(
        "--var-use-cov", action="store_true",
        help="Use covariance in varlss-eta fitting.")
    cont_group.add_argument(
        "--normalize-stacked-flux", action="store_true",
        help=("Force stacked flux to be one at the end."
              "Note this does not change STACKED_FLUX in attributes. "
              "It only updates CONT of each delta.")
    )
    cont_group.add_argument(
        "--eta-calib-ivar", action="store_true",
        help="Calibrate IVAR with eta estimates.")
    cont_group.add_argument(
        "--rfdwave", type=float, default=0.8,
        help="Rest-frame wave steps. Complies with forest limits")
    cont_group.add_argument(
        "--minimizer", default="iminuit", choices=["iminuit", "l_bfgs_b"],
        help="Minimizer to fit the continuum.")

    return parser


class PiccaContinuumFitter():
    """ Picca continuum fitter class.

    Fits spectra without coadding. Pipeline inverse variance preferably should
    be smoothed before fitting. Mean continuum and var_lss are smoothed using
    inverse weights and cubic spline to help numerical stability.

    When fitting for var_lss, number of wavelength bins in the observed frame
    for variance fitting ``nwbins`` are calculated by demanding 120 A steps
    between bins as closely as possible by :class:`VarLSSFitter`.

    Contruct an instance, then call :meth:`iterate` with local spectra.

    Parameters
    ----------
    args: argparse.Namespace
        Namespace. Wavelength values are taken to be the edges (not centers).
        See respective parsers for default values.

    Attributes
    ----------
    nbins: int
        Number of bins for the mean continuum in the rest frame.
    rfwave: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Rest-frame wavelength centers for the mean continuum.
    _denom: float
        Denominator for the slope term in the continuum model.
    meancont_interp: FastCubic1DInterp
        Fast cubic spline object for the mean continuum.
    comm: MPI.COMM_WORLD
        MPI comm object to reduce, broadcast etc.
    mpi_rank: int
        Rank of the MPI process.
    meanflux_interp: FastLinear1DInterp
        Interpolator for mean flux. If fiducial is not set, this equals to 1.
    flux_stacker: FluxStacker
        Stacks flux. Set up with 8 A wavelength bin size.
    varlss_fitter: VarLSSFitter or None
        None if fiducials are set for var_lss.
    varlss_interp: FastLinear1DInterp or FastCubic1DInterp
        Cubic spline for var_lss if fitting. Linear if from file.
    eta_interp: FastCubic1DInterp
        Interpolator for eta. Returns one if fiducial var_lss is set.
    niterations: int
        Number of iterations from ``args.num_iterations``.
    cont_order: int
        Order of continuum polynomial from ``args.cont_order``.
    outdir: str or None
        Directory to save catalogs. If None or empty, does not save.
    fit_eta: bool
        True if fitting eta and fiducial var_lss is not set.
    normalize_stacked_flux: bool
        Normalizes observed flux to be 1 if True.
    eta_calib_ivar: bool
        Calibrate IVAR with eta estimates.
    model: Derived(BaseContinuumModel)
        Continuum model to use
    """

    def _get_fiducial_interp(self, fname, col2read):
        """ Return an interpolator for mean flux or var_lss.

        FITS file must have a **STATS** extention, which must have **LAMBDA**,
        **MEANFLUX** and **VAR_LSS** columns. This is the same format as rawio
        output from picca, except **VAR** column in picca is the variance of
        flux not deltas. We break away from that convention by explicitly
        requiring variance on deltas in a new column. **LAMBDA** must be
        linearly and equally spaced. This function sets up ``col2read`` as
        FastLinear1DInterp object.

        Arguments
        ---------
        fname: str
            Filename of the FITS file.
        col2read: str
            Should be **MEANFLUX** or **VAR_LSS**.

        Returns
        -------
        FastLinear1DInterp

        Raises
        ------
        QsonicException
            If **LAMBDA** is not equally spaced or ``col2read`` is not in the
            file.
        """
        def _read(fname, col2read):
            with fitsio.FITS(fname) as fts:
                data = fts['STATS'].read()

            waves = data['LAMBDA']
            waves_0 = waves[0]
            dwave = waves[1] - waves[0]
            nsize = waves.size

            if not np.allclose(np.diff(waves), dwave):
                raise Exception(
                    "Failed to construct fiducial mean flux or varlss from "
                    f"{fname}::LAMBDA is not equally spaced.")

            data = np.array(data[col2read], dtype='d')

            return FastLinear1DInterp(waves_0, dwave, data, ep=np.zeros(nsize))

        return mpi_fnc_bcast(
            _read, self.comm, self.mpi_rank,
            f"Failed to construct fiducial mean flux or varlss from {fname}.",
            fname, col2read)

    def _set_meanflux_varlss_interps(self, args):
        if args.fiducial_meanflux:
            self.meanflux_interp = self._get_fiducial_interp(
                args.fiducial_meanflux, 'MEANFLUX')
        else:
            self.meanflux_interp = FastLinear1DInterp(
                args.wave1, args.wave2 - args.wave1, np.ones(3))

        if args.fiducial_varlss:
            self.varlss_fitter = None
            self.varlss_interp = self._get_fiducial_interp(
                args.fiducial_varlss, 'VAR_LSS')
        else:
            self.varlss_fitter = VarLSSFitter(
                args.wave1, args.wave2, use_cov=args.var_use_cov,
                comm=self.comm)
            self.varlss_interp = self.varlss_fitter.construct_interp(0.1)

    def _set_continuum_model(self, args):
        if args.continuum_model == "picca":
            from qsonic.continuum_models.picca_continuum_model import \
                PiccaContinuumModel

            self.model = PiccaContinuumModel(
                self.meancont_interp, self.meanflux_interp, self.varlss_interp,
                self.eta_interp, self.cont_order, self.rfwave[0], self._denom,
                args.minimizer)

        elif args.continuum_model == "true":
            from qsonic.continuum_models.true_continuum_model import \
                TrueContinuumModel

            self.model = TrueContinuumModel(
                self.meanflux_interp, self.varlss_interp)
            self.cont_order = 0
            self.niterations = 1

        elif args.continuum_model == "input":
            from qsonic.continuum_models.input_continuum_model import \
                InputContinuumModel

            self.model = InputContinuumModel(
                args.input_continuum_dir, self.meanflux_interp,
                self.varlss_interp, self.eta_interp)
            self.cont_order = 0

    def __init__(self, args):
        self.comm = MPI.COMM_WORLD
        self.mpi_rank = self.comm.Get_rank()

        self.args = args
        self.niterations = args.num_iterations
        self.cont_order = args.cont_order
        self.outdir = args.outdir
        self.fit_eta = args.var_fit_eta
        self.normalize_stacked_flux = args.normalize_stacked_flux
        self.eta_calib_ivar = args.eta_calib_ivar

        # We first decide how many bins will approximately satisfy
        # rest-frame wavelength spacing. Then we create wavelength edges, and
        # transform these edges into centers
        self.nbins = int(round(
            (args.forest_w2 - args.forest_w1) / args.rfdwave))
        edges, self.dwrf = np.linspace(
            args.forest_w1, args.forest_w2, self.nbins + 1, retstep=True)
        self.rfwave = (edges[1:] + edges[:-1]) / 2
        self._denom = np.log(self.rfwave[-1] / self.rfwave[0])

        self.meancont_interp = FastCubic1DInterp(
            self.rfwave[0], self.dwrf, np.ones(self.nbins),
            ep=np.zeros(self.nbins))

        self.flux_stacker = FluxStacker(
            args.wave1, args.wave2, 8., self.rfwave, self.dwrf, comm=self.comm)

        self._set_meanflux_varlss_interps(args)

        self.eta_interp = FastCubic1DInterp(
            self.varlss_interp.xp0, self.varlss_interp.dxp,
            np.ones_like(self.varlss_interp.fp),
            ep=np.zeros_like(self.varlss_interp.fp))

        self._set_continuum_model(args)

    def fit_continua(self, spectra_list):
        """ Fits all continua for a list of Spectrum objects.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects to fit.

        Raises
        ------
        QsonicException
            If there are no valid fits.
        """
        num_valid_fits = 0
        num_invalid_fits = 0

        # For each forest fit continuum
        for spec in spectra_list:
            self.model.fit_continuum(spec)

            if not spec.cont_params['valid']:
                num_invalid_fits += 1
            else:
                num_valid_fits += 1

        num_valid_fits = self.comm.allreduce(num_valid_fits)
        num_invalid_fits = self.comm.allreduce(num_invalid_fits)
        logging.info(f"Number of valid fits: {num_valid_fits}")
        logging.info(f"Number of invalid fits: {num_invalid_fits}")

        if num_valid_fits == 0:
            raise QsonicException("Crucial error: No valid continuum fits!")

        invalid_ratio = num_invalid_fits / (num_valid_fits + num_invalid_fits)
        if invalid_ratio > 0.2:
            logging.warning("More than 20% spectra have invalid fits.")

    def _project_normalize_meancont(self, new_meancont):
        """ Project out higher order Legendre polynomials from the new mean
        continuum since these are degenerate with the free fitting parameters.
        Returns a normalized mean continuum. Integrals are calculated using
        ``CubicSpline`` with ``ln lambda_RF`` as x array.

        Arguments
        ---------
        new_meancont: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            First estimate of the new mean continuum.

        Returns
        ---------
        new_meancont: :class:`ndarray <numpy.ndarray>`
            Legendere polynomials projected out and normalized mean continuum.
        mean: float
            Normalization of the mean continuum.
        """
        x = np.log(self.rfwave / self.rfwave[0]) / self._denom

        for ci in range(1, self.cont_order + 1):
            norm = 2 * ci + 1
            leg_ci = legendre(ci)(2 * x - 1)

            B = norm * CubicSpline(
                x, new_meancont * leg_ci, bc_type='natural', extrapolate=True
            ).integrate(0, 1)

            new_meancont -= B * leg_ci

        # normalize
        mean = CubicSpline(
            x, new_meancont, bc_type='natural', extrapolate=True
        ).integrate(0, 1)
        new_meancont /= mean

        return new_meancont, mean

    def update_mean_cont(self, spectra_list):
        """ Update the global mean continuum and stacked flux.

        Uses :attr:`forestivar_sm <qsonic.spectrum.Spectrum.forestivar_sm>`
        in inverse variance, but must be set beforehand.
        Raw mean continuum estimates are smoothed with a weighted
        :external+scipy:py:class:`scipy.interpolate.UnivariateSpline`. The mean
        continuum is removed from higher Legendre polynomials and normalized by
        the mean. This function updates
        :attr:`meancont_interp.fp <.meancont_interp>`.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects to fit.

        Returns
        ---------
        has_converged: bool
            True if all continuum updates on every point are less than 0.33
            times the error estimates.
        """
        self.flux_stacker.reset()

        self.model.stack_spectra(
            valid_spectra(spectra_list), self.flux_stacker)

        self.flux_stacker.calculate()
        w = self.flux_stacker.std_flux_rf > 0

        if not all(w):
            logging.warning("Extrapolating empty bins in the mean continuum.")

        norm_flux = self.flux_stacker.stacked_flux_rf.copy()
        norm_flux[~w] = np.mean(norm_flux[w])
        std_flux = self.flux_stacker.std_flux_rf.copy()
        std_flux[~w] = 10 * np.mean(std_flux[w])

        # Smooth new estimates
        spl = UnivariateSpline(self.rfwave, norm_flux, w=1 / std_flux)
        new_meancont = spl(self.rfwave)
        if self.model.stacks_residual_flux():
            new_meancont *= self.meancont_interp.fp

        # remove tilt and higher orders and normalize
        new_meancont, mean_ = self._project_normalize_meancont(new_meancont)

        norm_flux = new_meancont / self.meancont_interp.fp - 1
        std_flux /= mean_

        self.meancont_interp.reset(new_meancont, ep=std_flux)

        all_pt_test = np.all(np.abs(norm_flux) < 0.33 * std_flux)
        chi2_change = np.sum((norm_flux / std_flux)**2) / self.nbins
        has_converged = (chi2_change < 1e-3) | all_pt_test

        if self.mpi_rank != 0:
            return has_converged

        text = ("Continuum updates:\n" "rfwave\t| update\t| error\n")

        sl = np.s_[::max(1, self.nbins // 10)]
        for w, n, e in zip(self.rfwave[sl], norm_flux[sl], std_flux[sl]):
            text += f"{w:7.2f}\t| {n:7.2e}\t| +- {e:7.2e}\n"

        text += f"Change in chi2: {chi2_change*100:.4e}%"
        logging.info(text)

        return has_converged

    def update_var_lss_eta(self, spectra_list):
        """ Fit and update var_lss and eta if enabled. See
        :class:`VarLSSFitter` for fitting details.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects to fit.
        """
        self.varlss_fitter.reset()

        for spec in valid_spectra(spectra_list):
            for arm, wave_arm in spec.forestwave.items():
                cont = spec.cont_params['cont'][arm]
                delta = spec.forestflux[arm] / cont - 1
                ivar = spec.forestivar[arm] * cont**2

                self.varlss_fitter.add(wave_arm, delta, ivar)

        # Else, fit for var_lss
        if self.fit_eta:
            text = "Fitting var_lss and eta"
            initial_guess = np.vstack(
                (self.varlss_interp.fp, self.eta_interp.fp)).T
        else:
            text = "Fitting var_lss"
            initial_guess = self.varlss_interp.fp

        logging.info(text)
        y, ep = self.varlss_fitter.fit(initial_guess)

        if not self.fit_eta:
            self.varlss_interp.reset(y, ep=ep)
        else:
            self.varlss_interp.reset(y[:, 0], ep=ep[:, 0])
            self.eta_interp.reset(y[:, 1], ep=ep[:, 1])

        if self.mpi_rank != 0:
            return

        step = max(1, self.varlss_fitter.nwbins // 10)
        text = ("Variance fitting results:\n"
                "wave\t| var_lss +-  error \t|   eta   +-  error")

        for i in range(0, self.varlss_fitter.nwbins, step):
            w = self.varlss_fitter.waveobs[i]
            v = self.varlss_interp.fp[i]
            ve = self.varlss_interp.ep[i]
            n = self.eta_interp.fp[i]
            ne = self.eta_interp.ep[i]
            text += \
                f"\n{w:7.2f}\t| {v:7.2e} +- {ve:7.2e}\t| {n:7.2e} +- {ne:7.2e}"

        logging.info(text)

    def _normalize_flux(self, spectra_list):
        """Multiplies continuum estimates with stacked values in order to
        normalize flux in the observed grid to be 1 if
        ``--normalize-stacked-flux`` is passed.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects.
        """
        if not self.normalize_stacked_flux:
            return

        logging.info("Forcing stacked flux to be one.")

        for spec in valid_spectra(spectra_list):
            for arm, wave_arm in spec.forestwave.items():
                spec.cont_params['cont'][arm] *= self.flux_stacker(wave_arm)

    def _eta_calibate_ivar(self, spectra_list):
        """Divides forest ivar values by eta estimates if ``--eta-calib-ivar``
        is passed.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects.
        """
        if not self.eta_calib_ivar:
            return

        logging.info("Calibrating IVAR with eta.")

        for spec in spectra_list:
            for arm, wave_arm in spec.forestwave.items():
                eta = self.eta_interp(wave_arm)
                spec.forestivar[arm] /= eta

    def iterate(self, spectra_list):
        """Main function to fit continua and iterate.

        Consists of three major steps: initializing, fitting, updating global
        variables. The initialization sets ``cont_params`` variable of every
        Spectrum object. Continuum polynomial order is carried by setting
        ``cont_params[x]``. At each iteration:

        1. Global variables (mean continuum, var_lss) are saved to
           ``attributes.fits`` file. This ensures the order of what is used in
           each iteration.
        2. All spectra are fit.
        3. Mean continuum is updated by stacking, smoothing and removing
           degenarate modes. Check for convergence if update is small.
        4. If fitting for var_lss, fit and update by calculating variance
           statistics.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects to fit.
        """
        has_converged = False

        self.model.init_spectra(spectra_list)

        fname = f"{self.outdir}/attributes.fits" if self.outdir else ""
        fattr = MPISaver(fname, self.mpi_rank)

        for it in range(self.niterations):
            logging.info(f"Fitting iteration {it+1}/{self.niterations}")

            self.save(fattr, f"-{it + 1}")

            # Fit all continua one by one
            self.fit_continua(spectra_list)
            # Stack all spectra in each process
            # Broadcast and recalculate global functions
            has_converged = self.update_mean_cont(spectra_list)

            if self.varlss_fitter:
                self.update_var_lss_eta(spectra_list)

            if has_converged:
                logging.info("Iteration has converged.")
                break

        if not has_converged:
            logging.warning("Iteration has NOT converged.")

        self._normalize_flux(spectra_list)
        self._eta_calibate_ivar(spectra_list)

        self.save(fattr)
        if self.varlss_fitter is None:
            self.fit_final_varlss(spectra_list)
            self.save(fattr, '-fit')

        self.varlss_fitter.write(fattr)
        fattr.close()
        logging.info("All continua are fit.")

    def fit_final_varlss(self, spectra_list):
        logging.info("**These DO NOT go into weights**")
        self.varlss_fitter = VarLSSFitter(
            self.args.wave1, self.args.wave2, use_cov=self.args.var_use_cov,
            comm=self.comm)
        self.varlss_interp = self.varlss_fitter.construct_interp(0.1)
        self.eta_interp = self.varlss_fitter.construct_interp(1.0)

        self.update_var_lss_eta(spectra_list)
        self.update_mean_cont(spectra_list)
        logging.info("**These DO NOT go into weights**")

    def save(self, fattr, suff=''):
        """Save mean continuum and var_lss (if fitting) to a fits file.

        Arguments
        ---------
        fattr: MPISaver
            File handler to save only on master node.
        suff: str
            Suffix for the current iteration number.
        """
        fattr.write(
            [self.rfwave, self.meancont_interp.fp, self.meancont_interp.ep],
            names=['lambda_rf', 'mean_cont', 'e_mean_cont'],
            extname=f'CONT{suff}')

        fattr.write(
            [self.flux_stacker.waveobs, self.flux_stacker.stacked_flux],
            names=['lambda', 'stacked_flux'],
            extname=f'STACKED_FLUX{suff}')

        fattr.write(
            [self.rfwave, self.flux_stacker.stacked_flux_rf,
             self.flux_stacker.std_flux_rf],
            names=['lambda_rf', 'stacked_flux_rf', 'e_stacked_flux_rf'],
            extname=f'STACKED_FLUX_RF{suff}')

        if self.varlss_fitter is None:
            return

        fattr.write(
            [self.varlss_fitter.waveobs,
             self.varlss_interp.fp, self.varlss_interp.ep,
             self.eta_interp.fp, self.eta_interp.ep],
            names=['lambda', 'var_lss', 'e_var_lss', 'eta', 'e_eta'],
            extname=f'VAR_FUNC{suff}')

    def save_contchi2_catalog(self, spectra_list):
        """Save the chi2 catalog that includes information regarding chi2,
        mean_snr, targetid, etc. if ``self.outdir`` is set. All values are
        gathered to and saved on the master node.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects to fit.
        """
        if not self.outdir:
            return

        logging.info("Saving continuum chi2 catalog.")
        corder = self.cont_order + 1

        dtype = np.dtype([
            ('TARGETID', 'int64'), ('Z', 'f4'), ('HPXPIXEL', 'i8'),
            ('ARMS', 'U5'), ('MEANSNR', 'f4', 3), ('RSNR', 'f4'),
            ('MPI_RANK', 'i4'),
            ('CONT_valid', bool), ('CONT_chi2', 'f4'), ('CONT_dof', 'i4'),
            ('CONT_x', 'f4', (corder, )), ('CONT_xcov', 'f4', (corder**2, ))
        ])
        local_catalog = np.empty(len(spectra_list), dtype=dtype)

        for i, spec in enumerate(spectra_list):
            mean_snrs = np.zeros(3)
            _x = list(spec.mean_snr.values())
            mean_snrs[:len(_x)] = _x

            row = local_catalog[i]
            row['TARGETID'] = spec.targetid
            row['Z'] = spec.z_qso
            row['HPXPIXEL'] = spec.catrow['HPXPIXEL']
            row['ARMS'] = ",".join(spec.mean_snr.keys())
            row['MEANSNR'] = mean_snrs
            row['RSNR'] = spec.rsnr
            row['MPI_RANK'] = self.mpi_rank
            for lbl in ['valid', 'x', 'chi2', 'dof']:
                row[f'CONT_{lbl}'] = spec.cont_params[lbl]
            row['CONT_xcov'] = spec.cont_params['xcov'].ravel()

        all_catalogs = self.comm.gather(local_catalog)
        if self.mpi_rank == 0:
            all_catalog = np.concatenate(all_catalogs)
            fts = fitsio.FITS(
                f"{self.outdir}/continuum_chi2_catalog.fits", 'rw',
                clobber=True)
            fts.write(all_catalog, extname='CHI2_CAT')
            fts.close()


@njit("f8[:, :](i8[:], f8[:], f8[:], i8)")
def _fast_weighted_vector_bincount(x, delta, var, minlength):
    xvec = np.zeros((4, minlength), dtype=np.float_)
    y = delta**2
    y2 = y**2

    for i in range(x.size):
        xvec[0, x[i]] += delta[i]
        xvec[1, x[i]] += y[i]
        xvec[2, x[i]] += y2[i]
        xvec[3, x[i]] += var[i]

    return xvec


class VarLSSFitter():
    """ Variance fitter for the large-scale fluctuations.

    Input wavelengths and variances are the bin edges, so centers will be
    shifted. Valid bins require at least 500 pixels from 50 quasars. Assumes no
    spectra has `wave < w1obs` or `wave > w2obs`.

    .. note::

        This class is designed to be used in a linear fashion. You create it,
        add statistics to it and finally fit. After :meth:`fit` is called,
        :meth:`fit` and :meth:`add` **cannot** be called again. You may
        :meth:`reset` and start over.

    Usage::

        ...
        varfitter = VarLSSFitter(
            wave1, wave2, nwbins,
            var1, var2, nvarbins,
            nsubsamples=100, comm=comm)
        # Change static minimum numbers for valid statistics
        VarLSSFitter.min_num_pix = min_num_pix
        VarLSSFitter.min_num_qso = min_num_qso

        for delta in deltas_list:
            varfitter.add(delta.wave, delta.delta, delta.ivar)

        fit_results = np.ones((nwbins, 2))
        fit_results[:, 0] = 0.1
        fit_results, std_results = varfitter.fit(fit_results)

        varfitter.save("variance-file.fits")

        # You CANNOT call ``fit`` again!

    Parameters
    ----------
    w1obs: float
        Lower observed wavelength edge.
    w2obs: float
        Upper observed wavelength edge.
    nwbins: int, default: None
        Number of wavelength bins. If none, automatically calculated to yield
        120 A wavelength spacing.
    var1: float, default: 1e-4
        Lower variance edge.
    var2: float, default: 20
        Upper variance edge.
    nvarbins: int, default: 100
        Number of variance bins.
    nsubsamples: int, default: None
        Number of subsamples for the Jackknife covariance. Auto determined by
        number of bins (``nvarbins^2``) if None.
    use_cov: bool, default: False
        Use the Jackknife covariance when fitting.
    comm: MPI.COMM_WORLD or None, default: None
        MPI comm object to allreduce if enabled.

    Attributes
    ----------
    waveobs: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Wavelength centers in the observed frame.
    ivar_edges: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Inverse variance edges.
    minlength: int
        Minimum size of the combined bin count array. It includes underflow and
        overflow bins for both wavelength and variance bins.
    wvalid_bins: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Bool array slicer to get non-overflow bins of 1D arrays.
    subsampler: SubsampleCov
        Subsampler object that stores mean_delta in axis=0, var_delta in axis=1
        , var2_delta in axis=2, mean bin variance in axis=3.
    comm: MPI.COMM_WORLD or None, default: None
        MPI comm object to allreduce if enabled.
    mpi_rank: int
        Rank of the MPI process if ``comm!=None``. Zero otherwise.
    """
    min_num_pix = 500
    """int: Minimum number of pixels a bin must have to be valid."""
    min_num_qso = 50
    """int: Minimum number of quasars a bin must have to be valid."""

    @staticmethod
    def variance_function(var_pipe, var_lss, eta=1):
        """Variance model to be fit.

        .. math::

            \\sigma^2_\\mathrm{obs} = \\eta \\sigma^2_\\mathrm{pipe} +
            \\sigma^2_\\mathrm{LSS}

        Arguments
        ---------
        var_pipe: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Pipeline variance.
        var_lss: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Large-scale structure variance.
        eta: float
            Pipeline variance calibration scalar.

        Returns
        -------
        :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Expected variance of deltas.
        """
        return eta * var_pipe + var_lss

    def __init__(
            self, w1obs, w2obs, nwbins=None,
            var1=1e-4, var2=20., nvarbins=100,
            nsubsamples=None, use_cov=False, comm=None
    ):
        if nwbins is None:
            nwbins = int(round((w2obs - w1obs) / 120.))

        self.nwbins = nwbins
        self.nvarbins = nvarbins
        self.use_cov = use_cov
        self.comm = comm
        self._stats_calculated = False

        # Set up wavelength and inverse variance bins
        wave_edges, self.dwobs = np.linspace(
            w1obs, w2obs, nwbins + 1, retstep=True)
        self.waveobs = (wave_edges[1:] + wave_edges[:-1]) / 2
        self.ivar_edges = np.logspace(
            -np.log10(var2), -np.log10(var1), nvarbins + 1)

        # Set up arrays to store statistics
        self.minlength = (self.nvarbins + 2) * (self.nwbins + 2)
        # Bool array slicer for get non-overflow bins in 1D array
        self.wvalid_bins = np.zeros(self.minlength, dtype=bool)
        for iwave in range(self.nwbins):
            i1 = (iwave + 1) * (self.nvarbins + 2) + 1
            i2 = i1 + self.nvarbins
            wbinslice = np.s_[i1:i2]
            self.wvalid_bins[wbinslice] = True

        self._num_pixels = np.zeros(self.minlength, dtype=int)
        self._num_qso = np.zeros(self.minlength, dtype=int)

        # If ran with MPI, save mpi_rank first
        # Then shift each container to remove possibly over adding to 0th bin.
        if comm is not None:
            self.mpi_rank = comm.Get_rank()
            mpi_size = comm.Get_size()
        else:
            self.mpi_rank = 0
            mpi_size = 1

        if nsubsamples is None:
            nsubsamples = self.nvarbins**2
        istart = (self.mpi_rank * nsubsamples) // mpi_size
        # Axis 0 is mean
        # Axis 1 is var_delta
        # Axis 2 is var2_delta
        # Axis 3 is var_centers
        self.subsampler = SubsampleCov(
            (4, self.minlength), nsubsamples, istart)

    def construct_interp(self, y0=1.0):
        """Return a `FastCubic1DInterp` object that interpolates in observed
        wavelength bins.

        Arguments
        ---------
        y0: float, default: 1.0
            Initial constant value for the interpolator.

        Returns
        -------
        FastCubic1DInterp
        """
        return FastCubic1DInterp(
            self.waveobs[0], self.dwobs,
            y0 * np.ones(self.nwbins), ep=np.zeros(self.nwbins)
        )

    def reset(self):
        """Reset delta and num arrays to zero."""
        self.subsampler.reset()
        self._num_pixels.fill(0)
        self._num_qso.fill(0)
        self._stats_calculated = False

    def add(self, wave, delta, ivar):
        """Add statistics of a single spectrum. Updates delta and num arrays.

        Assumes no spectra has ``wave < w1obs`` or ``wave > w2obs``.

        Arguments
        ---------
        wave: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Wavelength array.
        delta: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Delta array.
        ivar: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Inverse variance array.
        """
        # add 1 to match searchsorted/bincount output/input
        wave_indx = np.clip(
            ((wave - self.waveobs[0]) / self.dwobs + 1.5).astype(int),
            0, self.nwbins + 1
        )
        ivar_indx = np.searchsorted(self.ivar_edges, ivar)
        all_indx = ivar_indx + wave_indx * (self.nvarbins + 2)
        var = np.zeros_like(ivar)
        w = ivar > 0
        var[w] = 1. / ivar[w]

        npix = np.bincount(all_indx, minlength=self.minlength)
        self._num_pixels += npix
        xvec = _fast_weighted_vector_bincount(
            all_indx, delta, var, self.minlength)
        self.subsampler.add_measurement(xvec, npix)

        npix[npix > 0] = 1
        self._num_qso += npix

    def calculate_subsampler_stats(self):
        """ Calculates mean, variance and error on the variance.

        It also calculates the delete-one Jackknife variance over
        ``nsubsamples``.
        The variance on var_delta is updated with the variance on the mean
        estimates, then it is regularized by calculated var2_delta
        (Gaussian estimates), where if Jackknife variance is smaller than
        the Gaussian estimate, it is replaced by the Gaussian estimate.

        Covariance is calculated if :attr:`use_cov` is set to True in init.
        Covariance of mean_delta^2 is propagated using the formula in
        :func:`qsonic.mathtools.block_covariance_of_square`.
        """
        if self._stats_calculated:
            return

        if self.comm is not None:
            self.subsampler.allreduce(self.comm, MPI.IN_PLACE)

            self.comm.Allreduce(MPI.IN_PLACE, self._num_pixels)
            self.comm.Allreduce(MPI.IN_PLACE, self._num_qso)

        self.subsampler.get_mean_n_var()
        if self.use_cov:
            self.subsampler.get_mean_n_cov(
                indices=[0, 1], blockdim=self.nvarbins + 2)

        m2 = self.subsampler.mean[0]**2
        self.subsampler.mean[1] -= m2
        self.subsampler.mean[2] -= self.subsampler.mean[1]**2

        w = self._num_pixels > 0
        self.subsampler.mean[2, w] /= self._num_pixels[w]

        # Update variance / covariance
        if self.use_cov:
            self.subsampler.covariance[1] += block_covariance_of_square(
                self.subsampler.mean[0],
                self.subsampler.variance[0],
                self.subsampler.covariance[0])
        else:
            self.subsampler.variance[1] += 4 * m2 * self.subsampler.variance[0]
            self.subsampler.variance[1] += 2 * self.subsampler.variance[0]**2
            # Regularized jackknife errors
            self.subsampler.variance[1] = np.where(
                self.subsampler.variance[1] > self.subsampler.mean[2],
                self.subsampler.variance[1],
                self.subsampler.mean[2]
            )

        self._stats_calculated = True

    def _smooth_fit_results(self, fit_results, std_results):
        w = fit_results > 0

        # Smooth new estimates
        if fit_results.ndim == 1:
            spl = UnivariateSpline(
                self.waveobs[w], fit_results[w], w=1 / std_results[w])

            fit_results = spl(self.waveobs)
        # else ndim == 2
        else:
            w = w[:, 0]
            spl1 = UnivariateSpline(
                self.waveobs[w], fit_results[w, 0], w=1 / std_results[w, 0])
            spl2 = UnivariateSpline(
                self.waveobs[w], fit_results[w, 1], w=1 / std_results[w, 1])

            fit_results[:, 0] = spl1(self.waveobs)
            fit_results[:, 1] = spl2(self.waveobs)

        return fit_results

    def _fit_array_shape_assert(self, arr):
        assert (arr.ndim == 1 or arr.ndim == 2)
        assert (arr.shape[0] == self.nwbins)
        if arr.ndim == 2:
            assert (arr.shape[1] == 2)

    def _get_wave_bin_slice_params(self, iwave):
        i1 = iwave * self.nvarbins
        i2 = i1 + self.nvarbins
        wave_slice = np.s_[i1:i2]
        w = ((self.num_pixels[wave_slice] >= VarLSSFitter.min_num_pix)
             & (self.num_qso[wave_slice] >= VarLSSFitter.min_num_qso))

        if w.sum() < 5:
            logging.warning(
                "Not enough statistics for VarLSSFitter at "
                f"{self.waveobs[iwave]:.2f} A. Increasing validity range...")
            w = ((self.num_pixels[wave_slice] >= VarLSSFitter.min_num_pix // 2)
                 & (self.num_qso[wave_slice] >= VarLSSFitter.min_num_qso // 2))

        if w.sum() < 5:
            logging.warning(
                "Still not enough statistics for VarLSSFitter at "
                f"{self.waveobs[iwave]:.2f} A.")
            return None, None, None

        if self.use_cov:
            err2use = self.cov_var_delta[iwave][:, w][w, :]

            try:
                _ = np.linalg.cholesky(err2use)
            except np.linalg.LinAlgError:
                logging.warning(
                    "Covariance matrix is not positive-definite "
                    f"{self.waveobs[iwave]:.2f} A. Fallback to diagonals.")

                err2use = self.e_var_delta[wave_slice][w]
        else:
            err2use = self.e_var_delta[wave_slice][w]

        return wave_slice, w, err2use

    def fit(self, initial_guess, smooth=True):
        """ Syncronize all MPI processes and fit for ``var_lss`` and ``eta``.

        Second axis always contains ``eta`` values. Example::

            var_lss = initial_guess[:, 0]
            eta = initial_guess[:, 1]

        This implemented using :func:`scipy.optimize.curve_fit` with
        :attr:`e_var_delta` or :attr:`cov_var_delta` as absolute errors. Domain
        is bounded to ``(0, 2)``. These fits are then smoothed via
        :external+scipy:py:class:`scipy.interpolate.UnivariateSpline`
        using weights from ``curve_fit``,
        while missing values or failed wavelength bins are extrapolated.

        Arguments
        ---------
        initial_guess: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Initial guess for var_lss and eta. If 1D array, eta is fixed to
            one. If 2D, its shape must be ``(nwbins, 2)``.
        smooth: bool, default: True
            Smooth results using UnivariateSpline.

        Returns
        ---------
        fit_results: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Smoothed fit results at observed wavelengths where missing values
            are extrapolated. 1D array containing LSS variance if
            ``initial_guess`` is 1D. 2D containing eta values on the second
            axis if ``initial_guess`` is 2D ndarray.
        std_results: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Error on ``var_lss`` from sqrt of ``curve_fit`` output. Same
            behavior as ``fit_results``.
        """
        self._fit_array_shape_assert(initial_guess)
        self.calculate_subsampler_stats()

        nfails = 0
        fit_results = np.zeros_like(initial_guess)
        std_results = np.zeros_like(initial_guess)

        for iwave in range(self.nwbins):
            wave_slice, w, err2use = self._get_wave_bin_slice_params(iwave)

            if w is None:
                nfails += 1
                continue

            try:
                pfit, pcov = curve_fit(
                    VarLSSFitter.variance_function,
                    self.var_centers[wave_slice][w],
                    self.var_delta[wave_slice][w],
                    p0=initial_guess[iwave],
                    sigma=err2use,
                    absolute_sigma=True,
                    check_finite=True,
                    bounds=(0, 2)
                )
            except Exception as e:
                nfails += 1
                logging.warning(
                    "VarLSSFitter failed at wave_obs: "
                    f"{self.waveobs[iwave]:.2f}. "
                    f"Reason: {e}. Extrapolating.")
            else:
                fit_results[iwave] = pfit
                std_results[iwave] = np.sqrt(np.diag(pcov))

        invalid_ratio = nfails / fit_results.shape[0]
        if invalid_ratio > 0.4:
            raise QsonicException(
                "VarLSSFitter failed at more than 40% points.")

        if nfails > 0:
            logging.warning(
                f"VarLSSFitter failed and extrapolated at {nfails} points.")

        # Smooth new estimates
        if smooth:
            fit_results = self._smooth_fit_results(fit_results, std_results)

        return fit_results, std_results

    def write(self, mpi_saver, min_snr=0, max_snr=100):
        """ Write variance statistics to FITS file in 'VAR_STATS' extension.

        Arguments
        ---------
        mpi_saver: MPISaver
            MPI FITS file handler.
        min_snr: float, default: 0
            Minimum SNR in this sample to be written into header.
        max_snr: float, default: 100
            Maximum SNR in this sample to be written into header.
        """
        hdr_dict = {
            'MINNPIX': VarLSSFitter.min_num_pix,
            'MINNQSO': VarLSSFitter.min_num_qso,
            'MINSNR': min_snr,
            'MAXSNR': max_snr,
            'WAVE1': self.waveobs[0],
            'WAVE2': self.waveobs[-1],
            'NWBINS': self.nwbins,
            'IVAR1': self.ivar_edges[0],
            'IVAR2': self.ivar_edges[-1],
            'NVARBINS': self.nvarbins
        }

        data_to_write = [
            np.repeat(self.waveobs, self.nvarbins),
            self.var_centers, self.e_var_centers,
            self.var_delta, self.e_var_delta,
            self.mean_delta, self.var2_delta, self.num_pixels, self.num_qso]

        names = ['wave', 'var_pipe', 'e_var_pipe', 'var_delta', 'e_var_delta',
                 'mean_delta', 'var2_delta', 'num_pixels', 'num_qso']

        if self.use_cov:
            cov_data = self.cov_var_delta.reshape(
                self.nwbins * self.nvarbins, self.nvarbins)
            data_to_write.append(cov_data)
            names.append("cov_var_delta")

        mpi_saver.write(
            data_to_write, names=names, extname="VAR_STATS", header=hdr_dict)

    @property
    def num_pixels(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`:
        Number of pixels in bins."""
        return self._num_pixels[self.wvalid_bins]

    @property
    def num_qso(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`:
        Number of quasars in bins."""
        return self._num_qso[self.wvalid_bins]

    @property
    def mean_delta(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`: Mean delta."""
        return self.subsampler.mean[0][self.wvalid_bins]

    @property
    def var_delta(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`:
        Variance delta."""
        return self.subsampler.mean[1][self.wvalid_bins]

    @property
    def e_var_delta(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`:
        Error on variance delta using delete-one Jackknife. The error of
        :attr:`mean_delta` is propagated, then regularized using
        :attr:`var2_delta`. See source code of
        :meth:`calculate_subsampler_stats`.
        """
        return np.sqrt(self.subsampler.variance[1][self.wvalid_bins])

    @property
    def cov_var_delta(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`:
        3D array of shape ``(nwbins, nvarbins, nvarbins)`` for the covariance
        on variance delta using delete-one Jackknife. The covariance
        of :attr:`mean_delta` is propagated. See
        :func:`qsonic.mathtools.block_covariance_of_square` for details.
        None if :attr:`use_cov` is False."""
        if not self.use_cov:
            return None
        cov = self.subsampler.covariance[1]
        return cov[1:-1, 1:-1, 1:-1]

    @property
    def var2_delta(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`:
        Variance delta^2."""
        return self.subsampler.mean[2][self.wvalid_bins]

    @property
    def var_centers(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`:
        Mean variance for the bin centers in **descending** order."""
        return self.subsampler.mean[3][self.wvalid_bins]

    @property
    def e_var_centers(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`:
        Error on the mean variance for the bin centers."""
        return np.sqrt(self.subsampler.variance[3][self.wvalid_bins])


class FluxStacker():
    """ The class to stack flux values to obtain IGM mean flux and other
    problems.

    This object can be called. Stacked flux is initialized to one. Reset before
    adding statistics.

    Parameters
    ----------
    w1obs: float
        Lower observed wavelength edge.
    w2obs: float
        Upper observed wavelength edge.
    dwobs: float
        Wavelength spacing.
    waverf: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Rest-frame wavelength bins.
    dwrf: float
        Wavelength spacing in the rest frame.
    comm: MPI.COMM_WORLD or None, default: None
        MPI comm object to allreduce if enabled.

    Attributes
    ----------
    waveobs: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Wavelength centers in the observed frame.
    nwbins: int
        Number of wavelength bins
    dwobs: float
        Wavelength spacing. Usually same as observed grid.
    nwrfbins: int
        Number of rest-frame wavelength bins
    waverf0: float
        The first rest-frame wavelength bin
    dwrf: float
        Rest-frame wavelength spacing.
    _interp: FastLinear1DInterp
        Interpolator. Saves stacked_flux in fp and weights in ep.
    _interp_rf: FastLinear1DInterp
        Rest-frame interpolator. Saves stacked_flux in fp and weights in ep.
    comm: MPI.COMM_WORLD or None, default: None
        MPI comm object to allreduce if enabled.
    """

    def __init__(self, w1obs, w2obs, dwobs, waverf, dwrf, comm=None):
        # Set up wavelength and inverse variance bins
        self.nwbins = int(round((w2obs - w1obs) / dwobs))
        wave_edges, self.dwobs = np.linspace(
            w1obs, w2obs, self.nwbins + 1, retstep=True)
        self.waveobs = (wave_edges[1:] + wave_edges[:-1]) / 2

        self.comm = comm

        self._interp = FastLinear1DInterp(
            self.waveobs[0], self.dwobs,
            np.ones(self.nwbins), ep=np.zeros(self.nwbins))

        self.nwrfbins = waverf.size
        self.waverf0, self.dwrf = waverf[0], dwrf
        self._interp_rf = FastLinear1DInterp(
            self.waverf0, self.dwrf, np.ones(self.nwrfbins),
            ep=np.zeros(self.nwrfbins))

    def __call__(self, wave):
        return self._interp(wave)

    def add(self, wave, wave_rf, weighted_flux, weighted_cont, weight):
        """ Add statistics of a single spectrum.

        Updates :attr:`stacked_flux` and :attr:`weights`. Assumes no spectra
        has ``wave < w1obs`` or ``wave > w2obs``.

        Arguments
        ---------
        wave: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Wavelength array in the observed frame.
        wave_rf: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Wavelength array in the rest frame.
        weighted_flux: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Weighted flux array. Specifically w * f/C.
        weighted_cont: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Weighted continuum array.
        weight: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Weight array.
        """
        wave_indx = ((wave - self.waveobs[0]) / self.dwobs + 0.5).astype(int)

        self._interp.fp += np.bincount(
            wave_indx, weights=weighted_flux, minlength=self.nwbins)
        self._interp.ep += np.bincount(
            wave_indx, weights=weight, minlength=self.nwbins)

        wave_indx = ((wave_rf - self.waverf0) / self.dwrf + 0.5).astype(int)

        self._interp_rf.fp += np.bincount(
            wave_indx, weights=weighted_cont, minlength=self.nwrfbins)
        self._interp_rf.ep += np.bincount(
            wave_indx, weights=weight, minlength=self.nwrfbins)

    def calculate(self):
        """Calculate stacked flux by allreducing if necessary."""
        if self.comm is not None:
            self.comm.Allreduce(MPI.IN_PLACE, self._interp.fp)
            self.comm.Allreduce(MPI.IN_PLACE, self._interp.ep)

            self.comm.Allreduce(MPI.IN_PLACE, self._interp_rf.fp)
            self.comm.Allreduce(MPI.IN_PLACE, self._interp_rf.ep)

        w = self._interp.ep > 0
        self._interp.fp[w] /= self._interp.ep[w]
        self._interp.fp[~w] = 0

        w = self._interp_rf.ep > 0
        self._interp_rf.fp[w] /= self._interp_rf.ep[w]
        self._interp_rf.fp[~w] = 0

        self._interp_rf.ep[w] = 1 / np.sqrt(self._interp_rf.ep[w])

    def reset(self):
        """Reset :attr:`stacked_flux` and :attr:`weights` arrays to zero."""
        self._interp.fp.fill(0)
        self._interp.ep.fill(0)

        self._interp_rf.fp.fill(0)
        self._interp_rf.ep.fill(0)

    @property
    def stacked_flux(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`: Stacked flux."""
        return self._interp.fp

    @property
    def weights(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`: Weights."""
        return self._interp.ep

    @property
    def stacked_flux_rf(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`: Stacked flux in
        the rest-frame."""
        return self._interp_rf.fp

    @property
    def std_flux_rf(self):
        """:external+numpy:py:class:`ndarray <numpy.ndarray>`: Error in
        the rest-frame."""
        return self._interp_rf.ep



class BaseContinuumModel():
    """The contract for BaseContinuumModel"""

    def __init__(self):
        raise NotImplementedError

    def stack_spectra(self, valid_spectra_list, flux_stacker):
        """Stacks spectra in the observed and rest-frame. Observed frame
        stacking is performed over f/C. Rest-frame stacking is performed over C
        only.

        Arguments
        ---------
        valid_spectra_list: list(Spectrum)
            Valid spectra objects to iterate.
        flux_stacker: FluxStacker
            Flux stacker object.
        """
        for spec in valid_spectra_list:
            for arm, wave_arm in spec.forestwave.items():
                wave_rf_arm = wave_arm / (1 + spec.z_qso)

                cont = spec.cont_params['cont'][arm]
                weight = spec.forestweight[arm] * cont**2
                weighted_flux = weight * spec.forestflux[arm] / cont
                weighted_cont = weight * cont

                flux_stacker.add(
                    wave_arm, wave_rf_arm, weighted_flux, weighted_cont, weight
                )

    def fit_continuum(self, spec):
        """Fits the continuum for a single Spectrum. Should modify
        :attr:`cont_params <qsonic.spectrum.Spectrum.cont_params>`
        dictionary's ``valid`` and ``cont`` keys at least. ``x, xcov, chi2``
        and ``dof`` keys should be set either here or in :meth:`init_spectra`.

        Arguments
        ---------
        spec: Spectrum
            Spectrum object to fit.
        """
        raise NotImplementedError

    def init_spectra(self, spectra_list):
        """ Initializes
        :attr:`cont_params <qsonic.spectrum.Spectrum.cont_params>` for a list
        of Spectrum objects. Specifically, set ``method`` key and any unused
        keys in ``x, xcov, chi2, dof``.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            List of Spectrum objects.
        """
        raise NotImplementedError

    def stacks_residual_flux(self):
        """Return ``True`` if the model stacks residual fluxes in the
        rest-frame in :meth:`stack_spectra` such as
        :class:`PiccaContinuumModel`. Otherwise, return ``False``.

        Returns
        -------
        bool
        """
        raise NotImplementedError



import pytest

import numpy as np
from scipy.interpolate import interp1d, CubicSpline
import numpy.testing as npt

import qsonic.mathtools


class TestMathtools(object):
    def test_FastLinear1DInterp(self):
        xin, dxp = np.linspace(320., 550., 5, retstep=True)
        fp = np.power(xin / 100. - 4., 3)

        fast_interp = qsonic.mathtools.FastLinear1DInterp(xin[0], dxp, fp)
        xarr = np.linspace(310., 560., 100)
        yarr = fast_interp(xarr)

        ytrue = interp1d(xin, fp, fill_value='extrapolate')(xarr)

        npt.assert_allclose(yarr, ytrue)

        xin, dxp = np.linspace(320., 550., 300, retstep=True)
        fp = np.power(xin / 100. - 4., 3)

        fast_interp = qsonic.mathtools.FastLinear1DInterp(xin[0], dxp, fp)
        xarr = np.linspace(310., 560., 100)
        yarr = fast_interp(xarr)

        ytrue = interp1d(xin, fp, fill_value='extrapolate')(xarr)

        npt.assert_allclose(yarr, ytrue)

    def test_FastCubic1DInterp(self):
        xin, dxp = np.linspace(320., 550., 300, retstep=True)
        fp = np.power(xin / 100. - 4., 3)

        fast_interp = qsonic.mathtools.FastCubic1DInterp(
            xin[0], dxp, fp, bc_type='natural')
        xarr = np.linspace(310., 560., 100)
        yarr = fast_interp(xarr)

        ytrue = CubicSpline(xin, fp, bc_type='natural', extrapolate=True)(xarr)

        npt.assert_allclose(yarr, ytrue)

        fast_interp = qsonic.mathtools.FastCubic1DInterp(
            xin[0], dxp, fp, bc_type='not-a-knot')
        xarr = np.linspace(310., 560., 100)
        yarr = fast_interp(xarr)

        ytrue = CubicSpline(
            xin, fp, bc_type='not-a-knot', extrapolate=True)(xarr)

        npt.assert_allclose(yarr, ytrue)

    def test_mypoly1d(self):
        coefs = np.array([5.5, 1.5, 0.7])
        xarr = np.linspace(-2, 2, 100)
        yarr = qsonic.mathtools.mypoly1d(coefs, xarr)

        ytrue = np.poly1d(coefs[::-1])(xarr)

        npt.assert_allclose(yarr, ytrue)

    def test_fft_gaussian_smooth(self):
        xarr = np.ones(2**10)
        xarr_sm = qsonic.mathtools.fft_gaussian_smooth(xarr)
        npt.assert_allclose(xarr, xarr_sm)

    def test_get_median_outlier_mask(self):
        rng = np.random.default_rng(10)
        size = 100
        sigmas = 1 + rng.poisson(5, size=size).astype(float)
        ivar = sigmas**-2
        flux = rng.normal(0, sigmas)

        w = qsonic.mathtools.get_median_outlier_mask(flux, ivar)
        assert not any(w)

        flux[20] = 1e3
        w = qsonic.mathtools.get_median_outlier_mask(flux, ivar)
        assert w[20]
        assert not any(w[:20]) and not any(w[21:])

        ivar[20:40] = 0

        w = qsonic.mathtools.get_median_outlier_mask(flux, ivar)
        assert all(w[20:40])
        assert not any(w[:20])
        assert not any(w[40:])

        # test different spectral shapes
        flux = rng.poisson(sigmas) - sigmas + (np.arange(size) / 25)**2
        ivar = sigmas**-2
        w = qsonic.mathtools.get_median_outlier_mask(flux, ivar)
        assert not any(w)

        flux = 100 + ((1 + np.arange(size)) / 5)**2
        ivar = flux**-2
        flux = rng.poisson(flux)
        assert not any(w)

    def test_get_smooth_ivar(self):
        ivar = np.ones(2**10)
        idces = np.array([16, 17, 18, 234, 235, 512, 667, 898, 910, 956])
        ivar[idces] = 0

        ivar_sm = qsonic.mathtools.get_smooth_ivar(ivar)
        npt.assert_allclose(ivar_sm[idces], 0)
        npt.assert_allclose(ivar_sm, ivar)

    def test_SubsampleCov_theory(self):
        n, m = 300000, 5
        nsample = 10000
        subsampler = qsonic.mathtools.SubsampleCov((2, m), nsample)

        randoms = np.random.default_rng(0).normal(loc=1, size=(n, m))

        true_mean = np.mean(randoms, axis=0)
        true_var = np.var(randoms, axis=0)
        var_on_mean = true_var / randoms.shape[0]
        rtol = 3.5 / np.sqrt(nsample - 1)

        for r in randoms:
            xvec = np.empty((2, m))
            xvec[0] = r
            xvec[1] = 2 * r
            subsampler.add_measurement(xvec, 1)

        mean, var = subsampler.get_mean_n_var(bias_correct=False)
        npt.assert_allclose(mean[0], true_mean)
        npt.assert_allclose(var[0], var_on_mean, rtol=rtol)
        npt.assert_allclose(var[1], 4 * var_on_mean, rtol=rtol)

        mean, cov = subsampler.get_mean_n_cov(bias_correct=False)
        npt.assert_allclose(mean[0], true_mean)
        npt.assert_allclose(mean[1], 2 * true_mean)
        npt.assert_allclose(cov[0].diagonal(), var_on_mean, rtol=rtol)
        npt.assert_allclose(cov[1].diagonal(), 4 * var_on_mean, rtol=rtol)

    def test_SubsampleCov_shape(self):
        subsampler = qsonic.mathtools.SubsampleCov((3, 10), 20)
        randoms = np.random.default_rng().normal(size=(100, 3, 10))
        randoms[:, 1, :] += 1
        randoms[:, 2, :] += 2

        true_mean = np.mean(randoms, axis=0)
        # true_var = np.std(randoms, axis=0)**2
        # var_on_mean = true_var / randoms.shape[0]

        for r in randoms:
            subsampler.add_measurement(r, 1)

        mean, var = subsampler.get_mean_n_var()
        npt.assert_allclose(mean, true_mean)
        npt.assert_equal(mean.shape, var.shape)

        mean, cov = subsampler.get_mean_n_cov()
        npt.assert_allclose(mean, true_mean)
        assert (len(cov) == 3)
        for jj in range(3):
            npt.assert_equal(cov[jj].shape, (10, 10))

        mean, cov = subsampler.get_mean_n_cov(indices=[0, 1], blockdim=5)
        npt.assert_allclose(mean, true_mean)
        assert (len(cov) == 3)
        for jj in range(2):
            npt.assert_equal(cov[jj].shape, (2, 5, 5))
        assert cov[2] is None

    def test_SubsampleCov_reset(self):
        subsampler = qsonic.mathtools.SubsampleCov((3, 10), 20)
        subsampler.reset()

        npt.assert_equal(subsampler.all_measurements.shape, (20, 3, 10))
        npt.assert_equal(subsampler.all_weights.shape, (20, 1, 10))


if __name__ == '__main__':
    pytest.main()



import logging
import numpy as np
from iminuit import Minuit
from scipy.optimize import minimize

from qsonic import QsonicException
from qsonic.mathtools import mypoly1d
from qsonic.continuum_models.base_continuum_model import BaseContinuumModel


class PiccaContinuumModel(BaseContinuumModel):
    """Picca continuum model class.

    Parameters
    ----------
    meancont_interp: FastCubic1DInterp
        Fast cubic spline object for the mean continuum.
    meanflux_interp: FastLinear1DInterp
        Interpolator for mean flux. If fiducial is not set, this equals to 1.
    varlss_interp: FastLinear1DInterp or FastCubic1DInterp
        Cubic spline for var_lss if fitting. Linear if from file.
    eta_interp: FastCubic1DInterp
        Interpolator for eta. Returns one if fiducial var_lss is set.
    cont_order: int
        Order of continuum polynomial from ``args.cont_order``.
    rfwave0: float
        First rest-frame wavelength center for the mean continuum.
    denom: float
        Denominator for the slope term in the continuum model.
    minimizer: str
        ``iminuit`` or ``l_bfgs_b`` to select the minimizer function.

    Attributes
    ----------
    minimizer: function
        Function that points to one of the minimizer options.
    """

    def __init__(
            self, meancont_interp, meanflux_interp, varlss_interp,
            eta_interp, cont_order, rfwave0, denom, minimizer
    ):
        self.meancont_interp = meancont_interp
        self.meanflux_interp = meanflux_interp
        self.varlss_interp = varlss_interp
        self.eta_interp = eta_interp
        self.cont_order = cont_order
        self.rfwave0 = rfwave0
        self.denom = denom

        if minimizer == "iminuit":
            self.minimizer = self._iminuit_minimizer
        elif minimizer == "l_bfgs_b":
            self.minimizer = self._scipy_l_bfgs_b_minimizer
        else:
            raise QsonicException(
                "Undefined minimizer. Developer forgot to implement.")

    def _continuum_costfn(self, x, wave, flux, ivar_sm, z_qso):
        """Cost function to minimize for each quasar.

        This is a modified chi2 where amplitude is also part of minimization.
        Cost of each arm is simply added to the total cost.

        Arguments
        ---------
        x: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Polynomial coefficients for quasar diversity.
        wave: dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
            Observed-frame wavelengths.
        flux: dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
            Flux.
        ivar_sm: dict(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
            Smooth inverse variance.
        z_qso: float
            Quasar redshift.

        Returns
        ---------
        cost: float
            Cost (modified chi2) for a given ``x``.
        """
        cost = 0

        for arm, wave_arm in wave.items():
            cont_est = self.get_continuum_model(x, wave_arm / (1 + z_qso))
            # no_neg = np.sum(cont_est<0)
            # penalty = wave_arm.size * no_neg**2

            cont_est *= self.meanflux_interp(wave_arm)

            var_lss = self.varlss_interp(wave_arm) * cont_est**2
            eta = self.eta_interp(wave_arm)
            weight = ivar_sm[arm] / (eta + ivar_sm[arm] * var_lss)
            w = weight > 0

            cost += np.dot(
                weight, (flux[arm] - cont_est)**2
            ) - np.log(weight[w]).sum()  # + penalty

        return cost

    def get_continuum_model(self, x, wave_rf_arm):
        """Returns interpolated continuum model.

        Arguments
        ---------
        x: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Polynomial coefficients for quasar diversity.
        wave_rf_arm: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Rest-frame wavelength per arm.

        Returns
        ---------
        cont: :external+numpy:py:class:`ndarray <numpy.ndarray>`
            Continuum at `wave_rf_arm` values given `x`.
        """
        slope = np.log(wave_rf_arm / self.rfwave0) / self.denom

        cont = self.meancont_interp(wave_rf_arm) * mypoly1d(x, 2 * slope - 1)
        # Multiply with resolution
        # Edges are difficult though

        return cont

    def _iminuit_minimizer(self, spec, a0):
        def _cost(x):
            return self._continuum_costfn(
                x, spec.forestwave, spec.forestflux, spec.forestivar_sm,
                spec.z_qso)

        x0 = np.zeros_like(spec.cont_params['x'])
        x0[0] = a0
        mini = Minuit(_cost, x0)
        mini.errordef = Minuit.LEAST_SQUARES
        mini.migrad()

        result = {}

        result['valid'] = mini.valid
        result['x'] = np.array(mini.values)
        result['xcov'] = np.array(mini.covariance)

        return result

    def _scipy_l_bfgs_b_minimizer(self, spec, a0):
        x0 = np.zeros_like(spec.cont_params['x'])
        x0[0] = a0
        mini = minimize(
            self._continuum_costfn,
            x0,
            args=(spec.forestwave,
                  spec.forestflux,
                  spec.forestivar_sm,
                  spec.z_qso),
            method='L-BFGS-B',
            bounds=None,
            jac=None
        )

        result = {}

        result['valid'] = mini.success
        result['x'] = mini.x
        result['xcov'] = mini.hess_inv.todense()

        return result

    def stack_spectra(self, valid_spectra_list, flux_stacker):
        """Stacks spectra in the observed and rest-frame. Observed-frame and
        rest-frame stacking is performed over residuals f/C.

        Arguments
        ---------
        valid_spectra_list: list(Spectrum)
            Valid spectra objects to iterate.
        flux_stacker: FluxStacker
            Flux stacker object.
        """
        for spec in valid_spectra_list:
            for arm, wave_arm in spec.forestwave.items():
                wave_rf_arm = wave_arm / (1 + spec.z_qso)

                cont = spec.cont_params['cont'][arm]
                weight = spec.forestweight[arm] * cont**2
                weighted_flux = weight * spec.forestflux[arm] / cont

                flux_stacker.add(
                    wave_arm, wave_rf_arm, weighted_flux, weighted_flux, weight
                )

    def fit_continuum(self, spec):
        """Fits the continuum for a single Spectrum.

        This function uses
        :attr:`forestivar_sm <qsonic.spectrum.Spectrum.forestivar_sm>` in
        inverse variance, which must be smoothed beforehand.
        It also modifies
        :attr:`cont_params <qsonic.spectrum.Spectrum.cont_params>`
        dictionary's ``valid, cont, x, xcov, chi2, dof`` keys.
        If the best-fitting continuum is **negative at any point**, the fit is
        **invalidated**. Chi2 is set separately without using the
        :meth:`cost function <._continuum_costfn>`.
        ``x`` key is the best-fitting parameter, and ``xcov`` is their inverse
        Hessian ``hess_inv`` given by
        :external+scipy:func:`scipy.optimize.minimize` using 'L-BFGS-B' method.

        Arguments
        ---------
        spec: Spectrum
            Spectrum object to fit.
        """
        # We can precalculate meanflux and varlss here,
        # and store them in respective keys to spec.cont_params

        def get_a0():
            a0 = 0
            n0 = 1e-6
            for arm, ivar_arm in spec.forestivar_sm.items():
                a0 += np.dot(spec.forestflux[arm], ivar_arm)
                n0 += np.sum(ivar_arm)

            return a0 / n0

        result = self.minimizer(spec, get_a0())
        spec.cont_params['valid'] = result['valid']

        if spec.cont_params['valid']:
            spec.cont_params['cont'] = {}
            for arm, wave_arm in spec.forestwave.items():
                cont_est = self.get_continuum_model(
                    result['x'], wave_arm / (1 + spec.z_qso))

                if any(cont_est < 0):
                    spec.cont_params['valid'] = False
                    break

                cont_est *= self.meanflux_interp(wave_arm)
                # cont_est *= self.flux_stacker(wave_arm)
                spec.cont_params['cont'][arm] = cont_est

        spec.set_forest_weight(self.varlss_interp, self.eta_interp)
        # We can further eliminate spectra based chi2
        spec.calc_continuum_chi2()

        if spec.cont_params['valid']:
            spec.cont_params['x'] = result['x']
            spec.cont_params['xcov'] = result['xcov']
        else:
            spec.cont_params['cont'] = None
            spec.cont_params['chi2'] = -1

    def init_spectra(self, spectra_list):
        """ Initializes
        :attr:`cont_params <qsonic.spectrum.Spectrum.cont_params>` for a list
        of Spectrum objects.

        Arguments
        ---------
        spectra_list: list(Spectrum)
            Spectrum objects to fit.
        """
        logging.info("Initializing Picca continuum fitting.")

        for spec in spectra_list:
            spec.cont_params['method'] = 'picca'
            spec.cont_params['x'] = np.append(
                spec.cont_params['x'][0], np.zeros(self.cont_order))
            spec.cont_params['xcov'] = np.eye(self.cont_order + 1)
            spec.cont_params['dof'] = \
                spec.get_real_size() - self.cont_order - 1

    def stacks_residual_flux(self):
        """:meth:`stack_spectra` stacks residual flux values f/C.

        Returns
        -------
        True: bool
        """
        return True



import logging

import fitsio
import numpy as np

from qsonic import QsonicException


def _logic_true(args):
    return True


def mpi_parse(
        parser, comm, mpi_rank, options=None, args_logic_fnc=_logic_true
):
    """ Parse arguments on the master node, then broadcast.

    Arguments
    ---------
    parser: argparse.ArgumentParser
        Parser to be used.
    comm: MPI.COMM_WORLD
        MPI comm object for bcast
    mpi_rank: int
        Rank of the MPI process.
    options: None or list, default: None
        Options to parse. None parses ``sys.argv``
    args_logic_fnc: Callable[[args], bool], default: True
        Logic function to check for args.
    """
    if mpi_rank == 0:
        try:
            args = parser.parse_args(options)
            if not args_logic_fnc(args):
                args = -1
        except SystemExit:
            args = -1
    else:
        args = -1

    args = comm.bcast(args)
    if args == -1:
        exit(0)

    return args


class _INVALID_VALUE(object):
    """ Sentinel for invalid values. """


def mpi_fnc_bcast(fnc, comm=None, mpi_rank=0, err_msg="", *args, **kwargs):
    """ Wrapper function to run function then broadcast. Return value of
    ``fnc`` must be broadcastable.

    Example::

        # Reading lines of integers
        # from file `filename_for_integers` using np.loadtxt
        import numpy as np

        ints_in_file = mpi_fnc_bcast(
                np.loadtxt,
                comm, mpi_rank, "Error while reading file.",
                filename_for_integers, dtype=int)

    Arguments
    ---------
    fnc: callable
        Function to evaluate.
    comm: MPI.COMM_WORLD or None, default: None
        MPI comm object for bcast
    mpi_rank: int, default: 0
        Rank of the MPI process.
    err_msg: str, default: ''
        Error message to use raising QsonicException.
    *args:
        Arguments such as ``fnc(3, 4)``.
    **kwargs:
        Keyword arguments such as ``fnc(dtype=int)``

    Returns
    -------
    result: fnc
        Function ``fnc(*args, **kwargs)``.

    Raises
    ------
    QsonicException
        If any error occur while doing ``fnc``.
    """
    result = _INVALID_VALUE
    if (mpi_rank == 0 or comm is None):
        try:
            result = fnc(*args, **kwargs)
        except Exception as e:
            logging.exception(e)
            logging.error(f"Error in {fnc.__module__}.{fnc.__name__}.")
            result = _INVALID_VALUE

    if comm is not None:
        result = comm.bcast(result)

    if result is _INVALID_VALUE:
        raise QsonicException(err_msg)

    return result


def balance_load(split_catalog, mpi_size):
    """Load balancing function. The return value can be scattered.

    Arguments
    ---------
    split_catalog: list(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        List of catalog. Each element is a ndarray with the same healpix.
    mpi_size: int
        Number of MPI tasks running.

    Returns
    ---------
    local_queue: list(list(:external+numpy:py:class:`ndarray <numpy.ndarray>`))
        Spectra that ranks are reponsible for in ``split_catalog`` format.
    """
    number_of_spectra = np.zeros(mpi_size, dtype=int)
    local_queues = [[] for _ in range(mpi_size)]
    split_catalog.sort(key=lambda x: x.size, reverse=True)  # Descending order
    for cat in split_catalog:
        min_idx = np.argmin(number_of_spectra)
        number_of_spectra[min_idx] += cat.size
        local_queues[min_idx].append(cat)

    return local_queues


class MPISaver():
    """ A simple class to write to a FITS file on master node.

    Parameters
    ----------
    fname: str
        Filename. Does not create if empty string.
    mpi_rank: int
        Rank of the MPI process. Creates FITS if 0.
    """

    def __init__(self, fname, mpi_rank):
        if mpi_rank == 0 and fname:
            self.fts = fitsio.FITS(fname, 'rw', clobber=True)
        else:
            self.fts = None

    def close(self):
        """ Close FITS file."""
        if self.fts is not None:
            self.fts.close()

    def write(self, data, names=None, extname=None, header=None):
        """ Write to FITS file.

        Arguments
        ---------
        data: list(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
            Data to write to extension.
        names: list(str)or None, default: None
            Column names for data.
        extname: str or None, default: None
            Extention name.
        header: dict or None, default: None
            Header dictionary to save.
        """
        if self.fts is not None:
            self.fts.write(data, names=names, extname=extname, header=header)



import pytest
from unittest import TestCase

import qsonic.scripts.qsonic_fit


class TestQsonicFit(TestCase):
    def test_parser(self):
        parser = qsonic.scripts.qsonic_fit.get_parser()

        options = (
            "--input-dir indir --catalog incat -o outdir "
            "--continuum-model true --mock-analysis --fiducial-meanflux mf "
            "--fiducial-varlss vs"
        ).split(' ')
        args = parser.parse_args(options)
        assert (args.input_dir == "indir")
        assert (args.catalog == "incat")
        assert (args.outdir == "outdir")

        x = qsonic.scripts.qsonic_fit.args_logic_fnc_qsonic_fit(args)
        assert x

        options = (
            "--input-dir indir --catalog incat -o outdir "
            "--true-continuum --fiducial-meanflux mf --fiducial-varlss vs"
        ).split(' ')
        args = parser.parse_args(options)
        x = qsonic.scripts.qsonic_fit.args_logic_fnc_qsonic_fit(args)
        assert not x

        options = (
            "--input-dir indir --catalog incat -o outdir "
            "--true-continuum --mock-analysis --fiducial-meanflux mf "
            "--fiducial-varlss vs --noise-calibration nc "
            "--varlss-as-additive-noise --flux-calibration fc"
        ).split(' ')
        args = parser.parse_args(options)
        x = qsonic.scripts.qsonic_fit.args_logic_fnc_qsonic_fit(args)
        assert x


if __name__ == '__main__':
    pytest.main()



""" This module contains functions to read, validate and make necessary changes
to DESI quasar catalogs. Can be imported without a need for MPI. """

import logging
import warnings

import fitsio
from healpy import ang2pix
import numpy as np
from numpy.lib.recfunctions import rename_fields, append_fields

from qsonic import QsonicException
from qsonic.mpi_utils import balance_load, mpi_fnc_bcast

_accepted_extnames = set(['QSO_CAT', 'ZCATALOG', 'METADATA'])
"""set: Accepted extensions for quasar catalog."""
_required_columns = [
    set(['TARGETID']), set(['Z']), set(['TARGET_RA', 'RA']),
    set(['TARGET_DEC', 'DEC'])
]
"""list(set): Required columns for all cases."""
_required_data_columns = [
    set(['SURVEY']),
    set(['COADD_LASTNIGHT', 'LAST_NIGHT', 'LASTNIGHT'])
]
"""list(set): Required columns for real data analysis."""
_required_tile_columns = [set(['TILEID']), set(['PETAL_LOC'])]
"""list(set): Required columns for tile data analysis."""
_optional_columns = [
    'HPXPIXEL', 'PROGRAM', 'VMIN_CIV_450', 'VMAX_CIV_450', 'VMIN_CIV_2000',
    'VMAX_CIV_2000'
]
"""list(str): Optional columns."""
_all_columns = [
    col for reqset in (
        _required_columns + _required_data_columns + _required_tile_columns
    ) for col in reqset
] + _optional_columns


def read_quasar_catalog(
        filename, is_mock=False, is_tile=False, keep_surveys=None,
        zmin=0, zmax=100.0
):
    """ Returns a quasar catalog object (ndarray).

    It is sorted in the following order:
    1. HPXPIXEL or TILEID
    2. SURVEY or PETAL_LOC (if applicable),
    3. TARGETID.
    BAL info included if available. It is required for BAL masking.
    If 'HPXPIXEL' column is not present, n_side is assumed 16 for mocks, 64 for
    data.

    Arguments
    ----------
    filename: str
        Filename to catalog.
    is_mock: bool, default: False
        If the catalog is for mocks.
    is_tile: bool, default: False
        If the catalog is for tiles. Duplicate TARGETIDs are allowed as long as
        they are in different TILEIDs.
    keep_surveys: None or list(str), default: None
        List of surveys to subselect. None keeps all.
    zmin: float, default: 0
        Minimum quasar redshift
    zmax: float, default: 100
        Maximum quasar redshift

    Returns
    ----------
    catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Sorted catalog.
    """
    n_side = 16 if is_mock else 64
    catalog = _read(filename)
    catalog = _validate_adjust_column_names(catalog, is_mock, is_tile)
    catalog = _prime_catalog(
        catalog, n_side, keep_surveys, zmin, zmax, is_tile)

    return catalog


def mpi_read_quasar_catalog(
        filename, comm=None, mpi_rank=0, is_mock=False, is_tile=False,
        keep_surveys=None, zmin=0, zmax=100
):
    """ Returns the same quasar catalog object on all MPI ranks.

    It is sorted in the following order: HPXPIXEL, SURVEY (if applicable),
    TARGETID. BAL info included if available. It is required for BAL masking.

    Can be used without MPI by passing ``comm=None`` and ``mpi_rank=0``.

    Arguments
    ----------
    filename: str
        Filename to catalog.
    comm: MPI comm object or None, default: None
        MPI comm object for bcast
    mpi_rank: int, default: 0
        Rank of the MPI process
    is_mock: bool, default: False
        If the catalog is for mocks.
    is_tile: bool, default: False
        If the catalog is for tiles. Duplicate TARGETIDs are allowed.
    keep_surveys: None or list(str), default: None
        List of surveys to subselect. None keeps all.
    zmin: float, default: 0
        Minimum quasar redshift
    zmax: float, default: 100
        Maximum quasar redshift

    Returns
    ----------
    catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Sorted catalog on all MPI ranks.

    Raises
    ------
    QsonicException
        If error occurs while reading the catalog.
    """
    catalog = mpi_fnc_bcast(
        read_quasar_catalog,
        comm, mpi_rank, "Error while reading catalog.",
        filename, is_mock, is_tile, keep_surveys, zmin, zmax)

    return catalog


def mpi_get_local_queue(
        filename, comm=None, mpi_rank=0, mpi_size=1, is_mock=False,
        is_tile=False, keep_surveys=None, zmin=0, zmax=100.0
):
    """Reads catalog on master and scatter a list of catalogs to every
    mpi_rank. If in tile format, sort key is 'TILEID'.

    Arguments
    ----------
    filename: str
        Filename to catalog.
    comm: MPI comm object or None, default: None
        MPI comm object for scatter
    mpi_rank: int
        Rank of the MPI process
    mpi_size: int
        Size of MPI processes
    is_mock: bool, default: False
        If the catalog is for mocks.
    is_tile: bool, default: False
        If the catalog is for tiles. Split key will be 'TILEID'. Duplicate
        TARGETIDs are allowed.
    keep_surveys: None or list(str), default: None
        List of surveys to subselect. None keeps all.
    zmin: float, default: 0
        Minimum quasar redshift
    zmax: float, default: 100
        Maximum quasar redshift

    Returns
    ----------
    local_queue: list(:external+numpy:py:class:`ndarray <numpy.ndarray>`)
        List of sorted catalogs.
    """
    # We decide forest filename list
    # Group into unique pixels

    if mpi_rank == 0:
        try:
            catalog = read_quasar_catalog(
                filename, is_mock, is_tile, keep_surveys, zmin, zmax)
            status = True
        except Exception as e:
            logging.exception(e)
            status = False
    else:
        status = False

    status = comm.bcast(status)
    if not status:
        raise QsonicException("Error while reading catalog.")

    if mpi_rank == 0:
        split_key = "TILEID" if is_tile else "HPXPIXEL"
        unique_pix, s = np.unique(catalog[split_key], return_index=True)
        split_catalog = np.split(catalog, s[1:])
        logging.info(
            f"There are {unique_pix.size} keys ({split_key})."
            " Don't use more MPI processes.")

        # Roughly equal number of spectra
        logging.info("Load balancing.")
        # Returns a list of catalog (ndarray)
        local_queue = balance_load(split_catalog, mpi_size)
    else:
        local_queue = None

    return comm.scatter(local_queue)


def _check_required_columns(required_cols, colnames):
    """Asserts all required columns are present.

    Arguments
    ----------
    required_cols: list(set)
        Required columns as list of sets.
    colnames: list(str)
        Present column names

    Raises
    ------
    Exception
        If none of a required set is in colnames.
    """
    for reqset in required_cols:
        if reqset.intersection(colnames):
            continue
        raise Exception(
            "One of these columns must be present in the catalog: "
            f"{', '.join(reqset)}!")


def _validate_adjust_column_names(catalog, is_mock, is_tile):
    """ Validate `catalog` for required columns in `_required_columns`.
    'SURVEY' is also required for data. 'TARGET_{RA, DEC}' is transformed to
    'RA' and 'DEC' in return.

    Arguments
    ----------
    catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog.
    is_mock: bool
        If the catalog is for mocks, does not perform 'SURVEY' check.
    is_tile: bool
        If the catalog is for tiles, duplicate TARGETIDs are allowed as long as
        they are in different TILEIDs.

    Returns
    ----------
    catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog. No checks are performed.
    """
    colnames = catalog.dtype.names
    # Check if required columns are present
    _check_required_columns(_required_columns, colnames)
    if not is_mock:
        _check_required_columns(_required_data_columns, colnames)
    if is_tile:
        _check_required_columns(_required_tile_columns, colnames)

    nqso = catalog.size
    logging.info(f"There are {nqso} quasars in the catalog.")

    if not is_tile and nqso != np.unique(catalog['TARGETID']).size:
        raise Exception("There are duplicate TARGETIDs in catalog!")

    if is_tile and nqso != np.unique(catalog[['TARGETID', 'TILEID']]).size:
        raise Exception(
            "There are duplicate TARGETID - TILEID combinations in catalog!")

    # Adjust column names
    colname_map = {}
    for x in ['RA', 'DEC']:
        if (f'TARGET_{x}' in colnames) and (x not in colnames):
            colname_map[f'TARGET_{x}'] = x
    if ('LASTNIGHT' not in colnames) and ('COADD_LASTNIGHT' in colnames):
        colname_map['COADD_LASTNIGHT'] = 'LASTNIGHT'
    elif ('LASTNIGHT' not in colnames) and ('LAST_NIGHT' in colnames):
        colname_map['LAST_NIGHT'] = 'LASTNIGHT'

    if colname_map:
        catalog = rename_fields(catalog, colname_map)

    return catalog


def _read(filename):
    """ Reads FITS file catalog with all columns present that are in
    `_all_columns`

    Arguments
    ----------
    filename: str

    Returns
    ----------
    catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog. No checks are performed.
    """
    logging.info(f'Reading catalogue from {filename}')
    fitsfile = fitsio.FITS(filename)
    extnames = [hdu.get_extname() for hdu in fitsfile]
    cat_hdu = _accepted_extnames.intersection(extnames)
    if not cat_hdu:
        cat_hdu = 1
        warnings.warn(
            "Catalog HDU not found by hduname. Using extension 1.",
            RuntimeWarning)
    else:
        cat_hdu = cat_hdu.pop()

    colnames = fitsfile[cat_hdu].get_colnames()
    keep_columns = [col for col in colnames if col in _all_columns]
    catalog = np.array(fitsfile[cat_hdu].read(columns=keep_columns))
    fitsfile.close()

    return catalog


def _add_healpix(catalog, n_side, keep_columns):
    """ Add 'HPXPIXEL' column to catalog if not present.

    Arguments
    ----------
    catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog.
    n_side: int
        Healpix nside.
    keep_columns: list(str)
        List of surveys to subselect.

    Returns
    ----------
    catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        'HPXPIXEL' calculated and added catalog.
    """
    if 'HPXPIXEL' not in keep_columns:
        pixnum = ang2pix(
            n_side, catalog['RA'], catalog['DEC'], lonlat=True, nest=True)
        catalog = append_fields(
            catalog, 'HPXPIXEL', pixnum, dtypes=int, usemask=False)

    return catalog


def _prime_catalog(
        catalog, n_side, keep_surveys, zmin, zmax, is_tile
):
    """ Returns quasar catalog object. It is sorted in the following order:
    sort_key (HPXPIXEL), SURVEY (if applicable), TARGETID

    Arguments
    ----------
    catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Catalog.
    n_side: int
        Healpix nside.
    keep_surveys: list(str)
        List of surveys to subselect.
    zmin: float
        Minimum quasar redshift
    zmax: float
        Maximum quasar redshift
    is_tile: bool
        If the catalog is for tiles. Sort oder will be TILEID, PETAL_LOC.

    Returns
    ----------
    catalog: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        Sorted catalog. BAL info included if available (req. for BAL masking)
    """
    colnames = catalog.dtype.names

    # Redshift cuts
    w = (catalog['Z'] >= zmin) & (catalog['Z'] <= zmax)
    logging.info(f"There are {w.sum()} quasars in the redshift range.")
    catalog = catalog[w]

    if is_tile:
        sort_order = ['TILEID', 'PETAL_LOC', 'TARGETID']
    else:
        sort_order = ["HPXPIXEL", 'TARGETID']

    # Filter all the objects in the catalogue not belonging to the specified
    # surveys.
    if 'SURVEY' in colnames and keep_surveys is not None:
        w = np.isin(catalog["SURVEY"], keep_surveys)
        catalog = catalog[w]
        if len(keep_surveys) > 1 and not is_tile:
            sort_order.insert(1, 'SURVEY')
        logging.info(
            f"There are {w.sum()} quasars in given surveys {keep_surveys}.")

    if catalog.size == 0:
        raise Exception("Empty quasar catalogue.")

    catalog = _add_healpix(catalog, n_side, colnames)
    catalog.sort(order=sort_order)

    return catalog



import argparse
import glob
import logging
import time
from os import makedirs as os_makedirs
import warnings

import numpy as np

from qsonic import QsonicException
import qsonic.catalog
import qsonic.io
from qsonic.masks import BALMask
from qsonic.mpi_utils import mpi_parse, mpi_fnc_bcast, MPISaver
from qsonic.picca_continuum import VarLSSFitter
from qsonic.spectrum import add_wave_region_parser


def get_parser(add_help=True):
    """Constructs the parser needed for the script.

    Arguments
    ---------
    add_help: bool, default: True
        Add help to parser.

    Returns
    -------
    parser: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(
        add_help=add_help,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog='Note: Not every option is used in this script.')

    iogroup = parser.add_argument_group(
        'Input/output parameters and selections')
    iogroup.add_argument(
        "--input-dir", '-i', required=True,
        help="Input directory.")
    iogroup.add_argument(
        "--outdir", '-o', required=True,
        help="Output directory to save files.")
    iogroup.add_argument("--fbase", help="Basename", default="qsonic-eta-fits")
    iogroup.add_argument(
        "--catalog",
        help="Catalog filename to enable catalog related removals.")
    iogroup.add_argument(
        "--mock-analysis", action="store_true",
        help="Input folder is mock. Uses nside=16")
    iogroup.add_argument(
        "--keep-surveys", nargs='*',
        help="Surveys to keep. Empty keeps all.")
    iogroup.add_argument(
        "--remove-bal-qsos", action="store_true",
        help="Removes BAL sightlines in the catalog option.")
    iogroup.add_argument(
        "--remove-targetid-list",
        help="Text file with TARGETIDs to exclude from analysis.")

    vargroup = parser.add_argument_group(
        'Variance fitting parameters')
    vargroup.add_argument(
        "--nvarbins", help="Number of variance bins (logarithmically spaced).",
        default=100, type=int)
    vargroup.add_argument(
        "--var-use-cov", action="store_true",
        help="Use covariance in varlss-eta fitting.")
    vargroup.add_argument(
        "--nwbins", default=None, type=int,
        help="Number of wavelength bins. None creates bins with 120 A spacing")
    vargroup.add_argument(
        "--var1", help="Lower variance bin.", default=1e-4, type=float)
    vargroup.add_argument(
        "--var2", help="Upper variance bin.", default=20., type=float)
    vargroup.add_argument(
        "--min-snr", help="Minimum SNR of the forest.",
        default=0, type=float)
    vargroup.add_argument(
        "--max-snr", help="Maximum SNR of the forest.",
        default=100, type=float)

    parser = add_wave_region_parser(parser)

    return parser


def mpi_set_targetid_list_to_remove(args, comm=None, mpi_rank=0):
    """ Return a ndarray of TARGETIDs to remove from the sample.

    Can be used without MPI by passing ``comm=None`` (which is the default.)

    Arguments
    ---------
    args: argparse.Namespace
        Options passed to script.
    comm: MPI.COMM_WORLD or None, default: None
        Communication object broadcast data.
    mpi_rank: int, default: 0
        Rank of the MPI process.

    Returns
    -------
    ids_to_remove: :external+numpy:py:class:`ndarray <numpy.ndarray>`
        TARGETIDs to remove from the sample.

    Raises
    ------
    QsonicException
        If error occurs while reading ``args.remove_targetid_list`` or if
        ``--remove_bal_qsos`` is passed but the input catalog is missing BAL
        columns .
    """
    ids_to_remove = np.array([], dtype=int)
    if args.remove_targetid_list:
        ids_to_remove = mpi_fnc_bcast(
            np.loadtxt, comm, mpi_rank,
            "Error while reading remove_targetid_list.",
            args.remove_targetid_list, dtype=int)

    if args.catalog:
        catalog = qsonic.catalog.mpi_read_quasar_catalog(
            args.catalog, comm, mpi_rank, args.mock_analysis)
    else:
        catalog = None

    if catalog is not None and args.remove_bal_qsos:
        logging.info("Checking BAL mask.")
        BALMask.check_catalog(catalog)

        sel_ai = (catalog['VMIN_CIV_450'] > 0) & (catalog['VMAX_CIV_450'] > 0)
        sel_bi = (
            (catalog['VMIN_CIV_2000'] > 0) & (catalog['VMAX_CIV_2000'] > 0))
        sel_bal = np.any(sel_ai, axis=1) | np.any(sel_bi, axis=1)
        bal_targetids = catalog['TARGETID'][sel_bal]
        ids_to_remove = np.concatenate((ids_to_remove, bal_targetids))
        logging.info(f"Removing {bal_targetids.size} BAL sightlines")

    if catalog is not None and args.keep_surveys:
        sel_survey = np.isin(catalog['SURVEY'], args.keep_surveys)
        remove_sur_tids = catalog['TARGETID'][~sel_survey]
        logging.info(f"Removing {remove_sur_tids.size} non survey sightlines")
        ids_to_remove = np.concatenate((ids_to_remove, remove_sur_tids))

    return ids_to_remove


def mpi_read_all_deltas(args, comm=None, mpi_rank=0, mpi_size=1):
    start_time = time.time()
    logging.info("Reading deltas.")

    all_delta_files = mpi_fnc_bcast(
        glob.glob, comm, mpi_rank,
        f"Delta files are not found in {args.input_dir}.",
        f"{args.input_dir}/delta-*.fits*")

    ndelta_all = len(all_delta_files)
    logging.info(f"There are {ndelta_all} delta files.")
    if mpi_size > ndelta_all:
        warnings.warn(
            "There are more MPI processes then number of delta files.")

    nfiles_per_rank = max(1, ndelta_all // mpi_size)
    i1 = nfiles_per_rank * mpi_rank
    i2 = min(ndelta_all, i1 + nfiles_per_rank)
    files_this_rank = all_delta_files[i1:i2]

    deltas_list = [qsonic.io.read_deltas(fname) for fname in files_this_rank]

    etime = (time.time() - start_time) / 60  # min
    logging.info(f"Rank{mpi_rank} read {i2-i1} deltas in {etime:.1f} mins.")

    return deltas_list


def mpi_stack_fluxes(args, comm, deltas_list):
    dwave = deltas_list[0].header['DELTA_LAMBDA']
    nwaveobs = int((args.wave2 - args.wave1) / dwave) + 1
    waveobs = np.linspace(args.wave1, args.wave2, nwaveobs)
    stacked_flux = np.zeros(nwaveobs)
    weights = np.zeros(nwaveobs)

    for delta in deltas_list:
        flux = 1 + delta.delta
        idx = np.round((delta.wave - args.wave1) / dwave).astype(int)
        w = (idx >= 0) & (idx < nwaveobs)
        stacked_flux[idx[w]] += flux[w] * delta.weight[w]
        weights[idx[w]] += delta.weight[w]

    # Save stacked_flux to buffer, then used stacked_flux to store reduced
    # weights. Place them properly in the end.
    buf = np.zeros(nwaveobs)
    comm.Allreduce(stacked_flux, buf)
    stacked_flux *= 0
    comm.Allreduce(weights, stacked_flux)
    weights = stacked_flux
    stacked_flux = buf

    w = weights > 0
    stacked_flux[w] /= weights[w]
    stacked_flux[~w] = 0

    return waveobs, stacked_flux


def mpi_run_all(comm, mpi_rank, mpi_size):
    args = mpi_parse(get_parser(), comm, mpi_rank)
    if mpi_rank == 0:
        os_makedirs(args.outdir, exist_ok=True)

    varfitter = VarLSSFitter(
        args.wave1, args.wave2, args.nwbins,
        args.var1, args.var2, args.nvarbins,
        use_cov=args.var_use_cov, comm=comm)

    ids_to_remove = mpi_set_targetid_list_to_remove(args, comm, mpi_rank)

    def _is_kept(delta):
        return (
            (delta.targetid not in ids_to_remove)
            and (delta.mean_snr > args.min_snr)
            and (delta.mean_snr < args.max_snr)
        )

    deltas_list = mpi_read_all_deltas(args, comm, mpi_rank, mpi_size)
    # Flatten this list of lists and remove quasars
    deltas_list = [x for alist in deltas_list for x in alist if _is_kept(x)]

    for delta in deltas_list:
        varfitter.add(delta.wave, delta.delta, delta.ivar)

    logging.info("Fitting var_lss and eta")
    fit_results = np.ones((varfitter.nwbins, 2))
    fit_results[:, 0] = 0.1
    fit_results, std_results = varfitter.fit(fit_results)

    # Save variance stats to file
    logging.info("Saving variance stats to files")
    suffix = f"snr{args.min_snr:.1f}-{args.max_snr:.1f}"
    tmpfilename = f"{args.outdir}/{args.fbase}-{suffix}-variance-stats.fits"
    mpi_saver = MPISaver(tmpfilename, mpi_rank)
    varfitter.write(mpi_saver, args.min_snr, args.max_snr)
    logging.info(f"Variance stats saved in {tmpfilename}.")

    # Save fits results as well
    mpi_saver.write([
        varfitter.waveobs, fit_results[:, 0], std_results[:, 0],
        fit_results[:, 1], std_results[:, 1]],
        names=['lambda', 'var_lss', 'e_var_lss', 'eta', 'e_eta'],
        extname="VAR_FUNC"
    )

    waveobs, stacked_flux = mpi_stack_fluxes(args, comm, deltas_list)
    mpi_saver.write(
        [waveobs, stacked_flux],
        names=["lambda", "stacked_flux"],
        extname="STACKED_FLUX")

    mpi_saver.close()


def main():
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    mpi_rank = comm.Get_rank()
    mpi_size = comm.Get_size()

    logging.basicConfig(
        format='%(asctime)s - %(levelname)s: %(message)s',
        datefmt='%Y/%m/%d %I:%M:%S %p',
        level=logging.DEBUG if mpi_rank == 0 else logging.CRITICAL)

    try:
        mpi_run_all(comm, mpi_rank, mpi_size)
    except QsonicException as e:
        logging.exception(e)
        exit(1)
    except Exception as e:
        logging.critical(
            f"Unexpected error on Rank{mpi_rank}: {e}. Abort.", exc_info=True)
        comm.Abort()
